import asyncio
import base64
import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Awaitable, Callable, Dict
from unittest import mock

import aiohttp
import pytest
from aiohttp import web
from aiohttp.test_utils import TestServer as _TestServer
from jose import jws
from yarl import URL

from neuro_sdk import (
    PASS_CONFIG_ENV_NAME,
    AuthError,
    Cluster,
    ConfigError,
    Factory,
    __version__,
)
from neuro_sdk._config import _AuthConfig, _AuthToken, _ConfigData
from neuro_sdk._login import JWT_STANDALONE_SECRET

from tests import _TestServerFactory


@pytest.fixture
def tmp_home(tmp_path: Path, monkeypatch: Any) -> Path:
    monkeypatch.setattr(Path, "home", lambda: tmp_path)  # Like as it's not enough
    if os.name == "nt" and sys.version_info >= (3, 8):
        monkeypatch.setenv("USERPROFILE", str(tmp_path))
    else:
        monkeypatch.setenv("HOME", str(tmp_path))

    return tmp_path


@pytest.fixture
def config_dir(
    tmp_home: Path, token: str, auth_config: _AuthConfig, cluster_config: Cluster
) -> Path:
    config_path = tmp_home / ".neuro"
    _create_config(config_path, token, auth_config, cluster_config)
    return config_path


def _create_config(
    nmrc_path: Path, token: str, auth_config: _AuthConfig, cluster_config: Cluster
) -> str:
    config = _ConfigData(
        auth_config=auth_config,
        auth_token=_AuthToken.create_non_expiring(token),
        url=URL("https://dev.neu.ro/api/v1"),
        admin_url=URL("https://dev.neu.ro/apis/admin/v1"),
        version=__version__,
        cluster_name=cluster_config.name,
        org_name=cluster_config.orgs[0],
        clusters={cluster_config.name: cluster_config},
    )
    Factory(nmrc_path)._save(config)
    assert nmrc_path.exists()
    return token


@dataclass
class MockForLoginControl:
    client_id: str = "banana"


@pytest.fixture
async def mock_for_login_factory(
    aiohttp_server: _TestServerFactory,
    token: str,
    unused_tcp_port_factory: Callable[[], int],
) -> Callable[[MockForLoginControl], Awaitable[_TestServer]]:
    async def _factory(
        control: MockForLoginControl, auth_enabled: bool = True
    ) -> _TestServer:
        callback_urls = [
            f"http://127.0.0.1:{unused_tcp_port_factory()}",
            f"http://127.0.0.1:{unused_tcp_port_factory()}",
            f"http://127.0.0.1:{unused_tcp_port_factory()}",
        ]

        async def config_handler(request: web.Request) -> web.Response:
            config_json: Dict[str, Any] = {
                "auth_url": str(srv.make_url("/authorize")),
                "token_url": str(srv.make_url("/oauth/token")),
                "logout_url": str(srv.make_url("/v2/logout")),
                "admin_url": str(srv.make_url("/apis/admin/v1")),
                "client_id": control.client_id,
                "audience": "https://test.dev.neu.ro",
                "headless_callback_url": str(srv.make_url("/oauth/show-code")),
                "callback_urls": callback_urls,
                "success_redirect_url": "http://example.com",
            }

            if not auth_enabled or (
                "Authorization" in request.headers
                and "incorrect" not in request.headers["Authorization"]
            ):
                cluster_config: Dict[str, Any] = {
                    "name": "default",
                    "registry_url": "https://registry-dev.test.com",
                    "storage_url": "https://storage-dev.test.com",
                    "users_url": "https://users-dev.test.com",
                    "monitoring_url": "https://monitoring-dev.test.com",
                    "secrets_url": "https://secrets-dev.test.com",
                    "disks_url": "https://disks-dev.test.com",
                    "buckets_url": "https://buckets-dev.test.com",
                    "resource_presets": [
                        {
                            "name": "gpu-small",
                            "credits_per_hour": "10",
                            "cpu": 7,
                            "memory_mb": 30 * 1024,
                            "gpu": 1,
                            "gpu_model": "nvidia-tesla-k80",
                        },
                        {
                            "name": "gpu-large",
                            "credits_per_hour": "10",
                            "cpu": 7,
                            "memory_mb": 60 * 1024,
                            "gpu": 1,
                            "gpu_model": "nvidia-tesla-v100",
                        },
                        {
                            "name": "cpu-small",
                            "credits_per_hour": "10",
                            "cpu": 2,
                            "memory_mb": 2 * 1024,
                        },
                        {
                            "name": "cpu-large",
                            "credits_per_hour": "10",
                            "cpu": 3,
                            "memory_mb": 14 * 1024,
                        },
                    ],
                }
                config_json.update(
                    {
                        "authorized": True,
                        "clusters": [
                            {**cluster_config, "name": "default"},
                            {**cluster_config, "name": "default2"},
                        ],
                    }
                )
            else:
                config_json["authorized"] = False
            return web.json_response(config_json)

        async def show_code(request: web.Request) -> web.Response:
            return web.json_response({})

        async def authorize(request: web.Request) -> web.Response:
            url = URL(request.query["redirect_uri"]).with_query(code="test_auth_code")
            raise web.HTTPSeeOther(location=url)

        async def new_token(request: web.Request) -> web.Response:
            return web.json_response(
                {"access_token": token, "expires_in": 3600, "refresh_token": token}
            )

        app = web.Application()
        app.router.add_get("/config", config_handler)
        app.router.add_get("/oauth/show-code", show_code)
        app.router.add_get("/authorize", authorize)
        app.router.add_post("/oauth/token", new_token)
        srv = await aiohttp_server(app)
        return srv

    return _factory


@pytest.fixture
async def mock_for_login(
    mock_for_login_factory: Callable[[MockForLoginControl], Awaitable[_TestServer]]
) -> _TestServer:
    return await mock_for_login_factory(MockForLoginControl())


class TestConfigFileInteraction:
    async def test_config_file_absent(self, tmp_home: Path) -> None:
        with pytest.raises(ConfigError, match=r"file.+not exists"):
            await Factory().get()

    async def test_config_dir_is_file(self, tmp_home: Path) -> None:
        Path(tmp_home / ".neuro").write_text("something")
        with pytest.raises(ConfigError, match=r"not a directory"):
            await Factory().get()

    async def test_config_file_is_dir(self, tmp_home: Path) -> None:
        path = Path(tmp_home / ".neuro")
        path.mkdir()
        (path / "db").mkdir()
        with pytest.raises(ConfigError, match=r"not a regular file"):
            await Factory().get()

    async def test_default_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        token = _create_config(tmp_home / ".neuro", token, auth_config, cluster_config)
        client = await Factory().get()
        await client.close()
        assert await client.config.token() == token

    async def test_preset_serialization(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        _create_config(tmp_home / ".neuro", token, auth_config, cluster_config)
        client = await Factory().get()
        await client.close()
        assert len(client.presets) > 0
        assert not client.presets["cpu-large"].scheduler_enabled
        assert not client.presets["cpu-large"].preemptible_node
        assert client.presets["cpu-large-p"].scheduler_enabled
        assert client.presets["cpu-large-p"].preemptible_node

    async def test_shorten_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        token = _create_config(
            tmp_home / "test.nmrc", token, auth_config, cluster_config
        )
        client = await Factory(Path("~/test.nmrc")).get()
        await client.close()
        assert await client.config.token() == token

    async def test_full_path(
        self,
        tmp_home: Path,
        token: str,
        auth_config: _AuthConfig,
        cluster_config: Cluster,
    ) -> None:
        config_path = tmp_home / "test.nmrc"
        token = _create_config(config_path, token, auth_config, cluster_config)
        client = await Factory(config_path).get()
        await client.close()
        assert await client.config.token() == token

    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="Windows does not supports UNIX-like permissions",
    )
    async def test_file_permissions(self, config_dir: Path) -> None:
        Path(config_dir).chmod(0o777)
        with pytest.raises(ConfigError, match=r"permission"):
            await Factory().get()


class TestLogin:
    async def show_dummy_browser(self, url: URL) -> None:
        async with aiohttp.ClientSession() as client:
            await client.get(url, allow_redirects=True)

    async def test_login_already_logged(self, config_dir: Path) -> None:
        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login(self.show_dummy_browser)

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        await Factory().login(self.show_dummy_browser, url=mock_for_login.make_url("/"))
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_login_to_server_without_auth(
        self,
        tmp_home: Path,
        mock_for_login_factory: Callable[..., Awaitable[_TestServer]],
    ) -> None:
        mock_for_login = await mock_for_login_factory(
            MockForLoginControl(), auth_enabled=False
        )
        await Factory().login(self.show_dummy_browser, url=mock_for_login.make_url("/"))
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

        client = await Factory(Path(nmrc_path)).get()
        await client.close()
        token = await client.config.token()

        assert client.config.username == "user"
        jws.verify(
            token, JWT_STANDALONE_SECRET, algorithms="HS256"
        )  # verify it is standalone token


class TestLoginWithToken:
    async def test_login_with_token_already_logged(self, config_dir: Path) -> None:
        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login_with_token(token="tokenstr")

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        await Factory().login_with_token(
            token="tokenstr", url=mock_for_login.make_url("/")
        )
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_incorrect_token(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        with pytest.raises(AuthError):
            await Factory().login_with_token(
                token="incorrect", url=mock_for_login.make_url("/")
            )
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file not written after login "


class TestLoginPassedConfig:
    @pytest.fixture()
    def make_conf_data(self, mock_for_login: _TestServer) -> Callable[[str], str]:
        def _make_config(token: str) -> str:
            data = {
                "token": token,
                "cluster": "default",
                "url": str(mock_for_login.make_url("/")),
            }
            return base64.b64encode(json.dumps(data).encode()).decode()

        return _make_config

    @pytest.fixture()
    def set_conf_to_env(
        self, monkeypatch: Any, make_conf_data: Callable[[str], str]
    ) -> Callable[[str], None]:
        def _set_env(token: str) -> None:
            config_data = make_conf_data(token)
            monkeypatch.setenv(PASS_CONFIG_ENV_NAME, config_data)

        return _set_env

    async def test_login_with_passed_config_already_logged(
        self, set_conf_to_env: Callable[[str], None], config_dir: Path
    ) -> None:
        set_conf_to_env("tokenstr")
        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login_with_passed_config()

    async def test_auto_login(
        self, tmp_home: Path, set_conf_to_env: Callable[[str], None]
    ) -> None:
        set_conf_to_env("tokenstr")
        client = await Factory().get()
        await client.close()
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_auto_login_fail(
        self, tmp_home: Path, set_conf_to_env: Callable[[str], None]
    ) -> None:
        set_conf_to_env("incorrect")
        with pytest.raises(AuthError):
            await Factory().get()
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file not written after login "

    async def test_normal_login(
        self, tmp_home: Path, set_conf_to_env: Callable[[str], None]
    ) -> None:
        set_conf_to_env("tokenstr")
        await Factory().login_with_passed_config()
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_normal_login_direct_token(
        self, tmp_home: Path, make_conf_data: Callable[[str], str]
    ) -> None:
        token_data = make_conf_data("tokenstr")
        await Factory().login_with_passed_config(token_data)
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "

    async def test_incorrect_token(
        self, tmp_home: Path, set_conf_to_env: Callable[[str], None]
    ) -> None:
        set_conf_to_env("incorrect")
        with pytest.raises(AuthError):
            await Factory().login_with_passed_config()
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file written after bad login "

    async def test_bad_data(
        self,
        tmp_home: Path,
        monkeypatch: Any,
    ) -> None:
        monkeypatch.setenv(PASS_CONFIG_ENV_NAME, "something")
        with pytest.raises(ConfigError):
            await Factory().login_with_passed_config()
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file written after bad login "

    async def test_no_data(
        self,
        tmp_home: Path,
        monkeypatch: Any,
    ) -> None:
        with pytest.raises(ConfigError):
            await Factory().login_with_passed_config()
        nmrc_path = tmp_home / ".neuro"
        assert not Path(nmrc_path).exists(), "Config file written after bad login "


class TestHeadlessLogin:
    async def test_login_headless_already_logged(self, config_dir: Path) -> None:
        async def get_auth_code_cb(url: URL) -> str:
            return ""

        with pytest.raises(ConfigError, match=r"already exists"):
            await Factory().login_headless(get_auth_code_cb)

    async def test_normal_login(
        self, tmp_home: Path, mock_for_login: _TestServer
    ) -> None:
        async def get_auth_code_cb(url: URL) -> str:
            assert url.with_query(None) == mock_for_login.make_url("/authorize")

            assert dict(url.query) == dict(
                response_type="code",
                code_challenge=mock.ANY,
                code_challenge_method="S256",
                client_id="banana",
                redirect_uri=str(mock_for_login.make_url("/oauth/show-code")),
                scope="offline_access",
                audience="https://test.dev.neu.ro",
            )
            return "test_code"

        await Factory().login_headless(
            get_auth_code_cb, url=mock_for_login.make_url("/")
        )
        nmrc_path = tmp_home / ".neuro"
        assert Path(nmrc_path).exists(), "Config file not written after login "


class TestLogout:
    async def test_logout_no_browser_callback(self, config_dir: Path) -> None:
        await Factory().logout()
        assert not config_dir.exists(), "Config not removed after logout\n" + "\n".join(
            p.name for p in config_dir.iterdir()
        )

    async def test_logout_with_browser_callback(
        self, auth_config: _AuthConfig, config_dir: Path
    ) -> None:
        async def show_browser(url: URL) -> None:
            expected_logout_url = auth_config.logout_url.with_query(
                client_id=auth_config.client_id
            )
            assert url == expected_logout_url

        await Factory().logout(show_browser)
        assert not config_dir.exists(), "Config not removed after logout\n" + "\n".join(
            p.name for p in config_dir.iterdir()
        )


class TestConfigRecovery:
    async def show_dummy_browser(self, url: URL) -> None:
        async with aiohttp.ClientSession() as client:
            await client.get(url, allow_redirects=True)

    async def test_recovery(
        self,
        tmp_home: Path,
        mock_for_login_factory: Callable[[MockForLoginControl], Awaitable[_TestServer]],
    ) -> None:
        control = MockForLoginControl(client_id="test1")
        mock_for_login = await mock_for_login_factory(control)
        await Factory().login(self.show_dummy_browser, url=mock_for_login.make_url("/"))
        with mock.patch("neuro_sdk.__version__", "21.13.13"):  # Impossible version
            control.client_id = "test2"
            client = await Factory().get()
            assert client.config._config_data.version == "21.13.13"
            await client.close()

        await mock_for_login.close()
        await asyncio.sleep(0.01)

    async def test_recovery_cluster_is_preserved(
        self,
        tmp_home: Path,
        mock_for_login_factory: Callable[[MockForLoginControl], Awaitable[_TestServer]],
    ) -> None:
        control = MockForLoginControl(client_id="test1")
        mock_for_login = await mock_for_login_factory(control)
        await Factory().login(self.show_dummy_browser, url=mock_for_login.make_url("/"))
        async with await Factory().get() as client:
            await client.config.switch_cluster("default2")

        with mock.patch("neuro_sdk.__version__", "21.13.13"):  # Impossible version
            control.client_id = "test2"
            client = await Factory().get()
            assert client.config.cluster_name == "default2"
            await client.close()

        await mock_for_login.close()
        await asyncio.sleep(0.01)



import base64
import contextlib
import json
import logging
import numbers
import os
import re
import sqlite3
import sys
import time
from dataclasses import dataclass, replace
from decimal import Decimal
from pathlib import Path
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Tuple, Union

import toml
from yarl import URL

from ._core import _Core
from ._errors import ConfigError
from ._login import AuthTokenClient, _AuthConfig, _AuthToken
from ._plugins import ConfigScope, PluginManager, _ParamType
from ._rewrite import rewrite_module
from ._server_cfg import Cluster, Preset, _ServerConfig, get_server_config
from ._utils import NoPublicConstructor, find_project_root, flat

WIN32 = sys.platform == "win32"
CMD_RE = re.compile("[A-Za-z][A-Za-z0-9-]*")

MALFORMED_CONFIG_MSG = "Malformed config. Please logout and login again."


SCHEMA = {
    "main": flat(
        """
        CREATE TABLE main (auth_config TEXT,
                           token TEXT,
                           expiration_time REAL,
                           refresh_token TEXT,
                           url TEXT,
                           admin_url TEXT,
                           version TEXT,
                           cluster_name TEXT,
                           org_name TEXT,
                           clusters TEXT,
                           timestamp REAL)"""
    )
}


logger = logging.getLogger(__package__)


@dataclass(frozen=True)
class _ConfigData:
    auth_config: _AuthConfig
    auth_token: _AuthToken
    url: URL
    admin_url: Optional[URL]
    version: str
    cluster_name: Optional[str]
    org_name: Optional[str]
    clusters: Mapping[str, Cluster]


@dataclass(frozen=True)
class _ConfigRecoveryData:
    url: URL
    cluster_name: str
    org_name: str
    refresh_token: str


@rewrite_module
class Config(metaclass=NoPublicConstructor):
    def __init__(self, core: _Core, path: Path, plugin_manager: PluginManager) -> None:
        self._core = core
        self._path = path
        self._plugin_manager = plugin_manager
        self.__config_data: Optional[_ConfigData] = None

    def _load(self) -> _ConfigData:
        ret = self.__config_data = _load(self._path)
        return ret

    @property
    def _config_data(self) -> _ConfigData:
        ret = self.__config_data
        if ret is None:
            return self._load()
        else:
            return ret

    @property
    def username(self) -> str:
        return self._config_data.auth_token.username

    @property
    def presets(self) -> Mapping[str, Preset]:
        return MappingProxyType(self._cluster.presets)

    @property
    def clusters(self) -> Mapping[str, Cluster]:
        return MappingProxyType(self._config_data.clusters)

    @property
    def cluster_name(self) -> str:
        if not self._config_data.clusters:
            raise RuntimeError(
                "There are no clusters available. Please logout and login again."
            )
        name = self._get_user_cluster_name()
        if name is None:
            name = self._config_data.cluster_name
        assert name
        return name

    def _get_user_cluster_name(self) -> Optional[str]:
        config = self._get_user_config()
        section = config.get("job")
        if section is not None:
            return section.get("cluster-name")
        return None

    @property
    def cluster_orgs(self) -> List[Optional[str]]:
        return self.clusters[self.cluster_name].orgs

    @property
    def org_name(self) -> Optional[str]:
        name = self._get_user_org_name()
        if name == "NO_ORG":
            return None
        if name is None:
            name = self._config_data.org_name
        return name

    def _get_user_org_name(self) -> Optional[str]:
        config = self._get_user_config()
        section = config.get("job")
        if section is not None:
            return section.get("org-name")
        return None

    @property
    def _cluster(self) -> Cluster:
        return self.get_cluster(self.cluster_name)

    def get_cluster(self, cluster_name: str) -> Cluster:
        try:
            return self._config_data.clusters[cluster_name]
        except KeyError:
            if self._get_user_cluster_name() is None:
                tip = "Please logout and login again."
            else:
                tip = "Please edit local user config file or logout and login again."
            raise RuntimeError(
                f"Cluster {cluster_name} doesn't exist in "
                f"a list of available clusters "
                f"{list(self._config_data.clusters)}. {tip}"
            ) from None

    async def _fetch_config(self) -> _ServerConfig:
        token = await self.token()
        return await get_server_config(self._core._session, self.api_url, token)

    async def check_server(self) -> None:
        from . import __version__

        if self._config_data.version != __version__:
            config_authorized = await self._fetch_config()
            if (
                config_authorized.clusters != self.clusters
                or config_authorized.auth_config != self._config_data.auth_config
            ):
                raise ConfigError(
                    "Neuro Platform CLI was updated. Please logout and login again."
                )
            self.__config_data = replace(self._config_data, version=__version__)
            _save(self._config_data, self._path)

    async def fetch(self) -> None:
        server_config = await self._fetch_config()
        if self.cluster_name not in server_config.clusters:
            # Raise exception here?
            # if yes there is not way to switch cluster without relogin
            raise RuntimeError(
                f"Cluster {self.cluster_name} doesn't exist in "
                f"a list of available clusters "
                f"{list(server_config.clusters)}. "
                f"Please logout and login again."
            )
        self.__config_data = replace(self._config_data, clusters=server_config.clusters)
        _save(self._config_data, self._path)

    async def switch_cluster(self, name: str) -> None:
        if self._get_user_cluster_name() is not None:
            raise RuntimeError(
                "Cannot switch the project cluster. "
                "Please edit the '.neuro.toml' file."
            )
        if name not in self.clusters:
            raise RuntimeError(
                f"Cluster {name} doesn't exist in "
                f"a list of available clusters {list(self.clusters)}. "
            )
        cluster_orgs = self.clusters[name].orgs
        org_name = self.org_name
        if org_name not in cluster_orgs:
            # Cannot keep using same org
            # If NO_ORG is available - use it
            # else use first in alphabetical order
            if None in cluster_orgs:
                org_name = None
            else:
                org_name = sorted(cluster_orgs, key=lambda it: it or "")[0]
        self.__config_data = replace(
            self._config_data, cluster_name=name, org_name=org_name
        )
        _save(self._config_data, self._path)

    async def switch_org(self, name: Optional[str]) -> None:
        if self._get_user_org_name() is not None:
            raise RuntimeError(
                "Cannot switch the project org. Please edit the '.neuro.toml' file."
            )
        if name not in self._cluster.orgs:
            cluster_org_names = [org or "NO_ORG" for org in self._cluster.orgs]
            raise RuntimeError(
                f"Org {name or 'NO_ORG'} doesn't exist in "
                f"a list of available orgs {list(cluster_org_names)} for "
                f"cluster '{self.cluster_name}'. "
            )
        self.__config_data = replace(self._config_data, org_name=name)
        _save(self._config_data, self._path)

    @property
    def api_url(self) -> URL:
        return self._config_data.url

    @property
    def admin_url(self) -> Optional[URL]:
        return self._config_data.admin_url

    @property
    def service_accounts_url(self) -> URL:
        # TODO: use URL returned from server when available
        return self._config_data.url / "service_accounts"

    @property
    def monitoring_url(self) -> URL:
        return self._cluster.monitoring_url

    @property
    def storage_url(self) -> URL:
        return self._cluster.storage_url

    @property
    def registry_url(self) -> URL:
        return self._cluster.registry_url

    @property
    def secrets_url(self) -> URL:
        return self._cluster.secrets_url

    @property
    def disk_api_url(self) -> URL:
        return self._cluster.disks_url

    @property
    def bucket_api_url(self) -> URL:
        return self._cluster.buckets_url

    async def token(self) -> str:
        token = self._config_data.auth_token
        if not token.is_expired():
            return token.token
        async with AuthTokenClient(
            self._core._session,
            url=self._config_data.auth_config.token_url,
            client_id=self._config_data.auth_config.client_id,
        ) as token_client:
            new_token = await token_client.refresh(token)
            self.__config_data = replace(self._config_data, auth_token=new_token)
            with self._open_db() as db:
                _save_auth_token(db, new_token)
            return new_token.token

    async def _api_auth(self) -> str:
        token = await self.token()
        return f"Bearer {token}"

    async def _docker_auth(self) -> Dict[str, str]:
        token = await self.token()
        return {"username": "token", "password": token}

    async def _registry_auth(self) -> str:
        token = await self.token()
        return "Basic " + base64.b64encode(
            f"{self.username}:{token}".encode("ascii")
        ).decode("ascii")

    async def get_user_config(self) -> Mapping[str, Any]:
        return _load_user_config(self._plugin_manager, self._path)

    def _get_user_config(self) -> Mapping[str, Any]:
        return _load_user_config(self._plugin_manager, self._path)

    @contextlib.contextmanager
    def _open_db(self, suppress_errors: bool = True) -> Iterator[sqlite3.Connection]:
        with _open_db_rw(self._path, suppress_errors) as db:
            yield db


def _load_user_config(plugin_manager: PluginManager, path: Path) -> Mapping[str, Any]:
    # TODO: search in several locations (HOME+curdir),
    # merge found configs
    filename = path / "user.toml"
    if not filename.exists():
        # Empty global configuration
        config: Mapping[str, Any] = {}
    elif not filename.is_file():
        raise ConfigError(f"User config {filename} should be a regular file")
    else:
        config = _load_file(
            plugin_manager, filename, allow_cluster_name=False, allow_org_name=False
        )
    try:
        project_root = find_project_root()
    except ConfigError:
        return config
    else:
        filename = project_root / ".neuro.toml"
        local_config = _load_file(
            plugin_manager, filename, allow_cluster_name=True, allow_org_name=True
        )
        return _merge_user_configs(config, local_config)


@contextlib.contextmanager
def _open_db_rw(
    path: Path, suppress_errors: bool = True
) -> Iterator[sqlite3.Connection]:
    path.mkdir(0o700, parents=True, exist_ok=True)  # atomically set proper bits
    path.chmod(0o700)  # fix security if config folder already exists

    config_file = path / "db"
    with sqlite3.connect(str(config_file)) as db:
        # forbid access to other users
        os.chmod(config_file, 0o600)

        db.row_factory = sqlite3.Row
        try:
            db.execute("PRAGMA journal_mode=WAL")
            yield db
        except sqlite3.DatabaseError as exc:
            if not suppress_errors:
                raise
            msg = "Cannot send the usage statistics: %s"
            if str(exc) != "database is locked":
                logger.warning(msg, repr(exc))
            else:
                logger.debug(msg, repr(exc))


@contextlib.contextmanager
def _open_db_ro(
    path: Path, *, skip_schema_check: bool = False
) -> Iterator[sqlite3.Connection]:
    config_file = path / "db"
    if not path.exists():
        raise ConfigError(f"Config at {path} does not exists. Please login.")
    if not path.is_dir():
        raise ConfigError(
            f"Config at {path} is not a directory. Please logout and login again."
        )
    if not config_file.is_file():
        raise ConfigError(
            f"Config {config_file} is not a regular file. "
            "Please logout and login again."
        )

    if not WIN32:
        stat_dir = path.stat()
        if stat_dir.st_mode & 0o777 != 0o700:
            raise ConfigError(
                f"Config {path} has compromised permission bits, "
                f"run 'chmod 700 {path}' first"
            )
        stat_file = config_file.stat()
        if stat_file.st_mode & 0o777 != 0o600:
            raise ConfigError(
                f"Config at {config_file} has compromised permission bits, "
                f"run 'chmod 600 {config_file}' first"
            )

    with sqlite3.connect(str(config_file)) as db:
        # forbid access for other users
        os.chmod(config_file, 0o600)

        if not skip_schema_check:
            _check_db(db)
        db.row_factory = sqlite3.Row
        yield db


def _load(path: Path) -> _ConfigData:
    try:
        with _open_db_ro(path) as db:
            cur = db.cursor()
            # only one row is always present normally
            cur.execute(
                """
                SELECT auth_config, token, expiration_time, refresh_token,
                       url, admin_url, version, cluster_name, org_name, clusters
                FROM main ORDER BY timestamp DESC LIMIT 1"""
            )
            payload = cur.fetchone()

        api_url = URL(payload["url"])
        if not payload["admin_url"]:
            admin_url = None
        else:
            admin_url = URL(payload["admin_url"])
        auth_config = _deserialize_auth_config(payload)
        clusters = _deserialize_clusters(payload)
        version = payload["version"]
        cluster_name = payload["cluster_name"]
        org_name = payload["org_name"]

        auth_token = _AuthToken(
            payload["token"], payload["expiration_time"], payload["refresh_token"]
        )

        return _ConfigData(
            auth_config=auth_config,
            auth_token=auth_token,
            url=api_url,
            admin_url=admin_url,
            version=version,
            cluster_name=cluster_name,
            org_name=org_name,
            clusters=clusters,
        )
    except (AttributeError, KeyError, TypeError, ValueError, sqlite3.DatabaseError):
        raise ConfigError(MALFORMED_CONFIG_MSG)


def _load_recovery_data(path: Path) -> _ConfigRecoveryData:
    try:
        with _open_db_ro(path, skip_schema_check=True) as db:
            cur = db.cursor()
            # only one row is always present normally
            try:
                cur.execute(
                    """
                    SELECT refresh_token, url, cluster_name, org_name
                    FROM main ORDER BY timestamp DESC LIMIT 1"""
                )
                payload = cur.fetchone()
            except sqlite3.OperationalError:
                # Maybe this config was created before org_name was added?
                cur.execute(
                    """
                    SELECT refresh_token, url, cluster_name
                    FROM main ORDER BY timestamp DESC LIMIT 1"""
                )
                payload = cur.fetchone()

        return _ConfigRecoveryData(
            url=URL(payload["url"]),
            cluster_name=payload["cluster_name"],
            refresh_token=payload["refresh_token"],
            org_name=payload["org_name"] if "org_name" in payload else None,
        )
    except (AttributeError, KeyError, TypeError, ValueError, sqlite3.DatabaseError):
        raise ConfigError(MALFORMED_CONFIG_MSG)


def _deserialize_auth_config(payload: Dict[str, Any]) -> _AuthConfig:
    auth_config = json.loads(payload["auth_config"])
    success_redirect_url = auth_config.get("success_redirect_url")
    if success_redirect_url:
        success_redirect_url = URL(success_redirect_url)
    return _AuthConfig(
        auth_url=URL(auth_config["auth_url"]),
        token_url=URL(auth_config["token_url"]),
        logout_url=URL(auth_config["logout_url"]),
        client_id=auth_config["client_id"],
        audience=auth_config["audience"],
        headless_callback_url=URL(auth_config["headless_callback_url"]),
        success_redirect_url=success_redirect_url,
        callback_urls=tuple(URL(u) for u in auth_config.get("callback_urls", [])),
    )


def _deserialize_clusters(payload: Dict[str, Any]) -> Dict[str, Cluster]:
    clusters = json.loads(payload["clusters"])
    ret: Dict[str, Cluster] = {}
    for cluster_config in clusters:
        cluster = Cluster(
            name=cluster_config["name"],
            orgs=cluster_config.get("orgs", [None]),
            registry_url=URL(cluster_config["registry_url"]),
            storage_url=URL(cluster_config["storage_url"]),
            users_url=URL(cluster_config["users_url"]),
            monitoring_url=URL(cluster_config["monitoring_url"]),
            secrets_url=URL(cluster_config["secrets_url"]),
            disks_url=URL(cluster_config["disks_url"]),
            buckets_url=URL(cluster_config["buckets_url"]),
            presets=dict(
                _deserialize_resource_preset(data)
                for data in cluster_config.get("presets", [])
            ),
        )
        ret[cluster.name] = cluster
    return ret


def _deserialize_resource_preset(payload: Dict[str, Any]) -> Tuple[str, Preset]:
    return (
        payload["name"],
        Preset(
            credits_per_hour=Decimal(payload["credits_per_hour"]),
            cpu=payload["cpu"],
            memory_mb=payload["memory_mb"],
            gpu=payload.get("gpu"),
            gpu_model=payload.get("gpu_model"),
            tpu_type=payload.get("tpu_type", None),
            tpu_software_version=payload.get("tpu_software_version", None),
            scheduler_enabled=payload.get("scheduler_enabled", False),
            preemptible_node=payload.get("preemptible_node", False),
        ),
    )


def _deserialize_auth_token(payload: Dict[str, Any]) -> _AuthToken:
    auth_payload = payload["auth_token"]
    return _AuthToken(
        token=auth_payload["token"],
        expiration_time=auth_payload["expiration_time"],
        refresh_token=auth_payload["refresh_token"],
    )


def _save_auth_token(db: sqlite3.Connection, token: _AuthToken) -> None:
    db.execute(
        "UPDATE main SET token=?, expiration_time=?, refresh_token=?",
        (token.token, token.expiration_time, token.refresh_token),
    )
    with contextlib.suppress(sqlite3.OperationalError):
        db.commit()


def _save(config: _ConfigData, path: Path, suppress_errors: bool = True) -> None:
    # The wierd method signature is required for communicating with existing
    # Factory._save()
    try:
        url = str(config.url)
        if not config.admin_url:
            admin_url = None
        else:
            admin_url = str(config.admin_url)
        auth_config = _serialize_auth_config(config.auth_config)
        clusters = _serialize_clusters(config.clusters)
        version = config.version
        cluster_name = config.cluster_name
        org_name = config.org_name
        token = config.auth_token
    except (AttributeError, KeyError, TypeError, ValueError):
        raise ConfigError(MALFORMED_CONFIG_MSG)

    with _open_db_rw(path, suppress_errors) as db:
        _init_db_maybe(db)

        cur = db.cursor()
        cur.execute("DELETE FROM main")
        cur.execute(
            """
            INSERT INTO main
            (auth_config, token, expiration_time, refresh_token,
             url, admin_url, version, cluster_name, org_name, clusters, timestamp)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (
                auth_config,
                token.token,
                token.expiration_time,
                token.refresh_token,
                url,
                admin_url,
                version,
                cluster_name,
                org_name,
                clusters,
                time.time(),
            ),
        )
        db.commit()


def _serialize_auth_config(auth_config: _AuthConfig) -> str:
    success_redirect_url = None
    if auth_config.success_redirect_url:
        success_redirect_url = str(auth_config.success_redirect_url)
    return json.dumps(
        {
            "auth_url": str(auth_config.auth_url),
            "token_url": str(auth_config.token_url),
            "logout_url": str(auth_config.logout_url),
            "client_id": auth_config.client_id,
            "audience": auth_config.audience,
            "headless_callback_url": str(auth_config.headless_callback_url),
            "success_redirect_url": success_redirect_url,
            "callback_urls": [str(u) for u in auth_config.callback_urls],
        }
    )


def _serialize_clusters(clusters: Mapping[str, Cluster]) -> str:
    ret: List[Dict[str, Any]] = []
    for cluster in clusters.values():
        cluster_config = {
            "name": cluster.name,
            "orgs": cluster.orgs,
            "registry_url": str(cluster.registry_url),
            "storage_url": str(cluster.storage_url),
            "users_url": str(cluster.users_url),
            "monitoring_url": str(cluster.monitoring_url),
            "secrets_url": str(cluster.secrets_url),
            "disks_url": str(cluster.disks_url),
            "buckets_url": str(cluster.buckets_url),
            "presets": [
                _serialize_resource_preset(name, preset)
                for name, preset in cluster.presets.items()
            ],
        }
        ret.append(cluster_config)
    return json.dumps(ret)


def _serialize_resource_preset(name: str, preset: Preset) -> Dict[str, Any]:
    return {
        "name": name,
        "credits_per_hour": str(preset.credits_per_hour),
        "cpu": preset.cpu,
        "memory_mb": preset.memory_mb,
        "gpu": preset.gpu,
        "gpu_model": preset.gpu_model,
        "tpu_type": preset.tpu_type,
        "tpu_software_version": preset.tpu_software_version,
        "scheduler_enabled": preset.scheduler_enabled,
        "preemptible_node": preset.preemptible_node,
    }


def _merge_user_configs(
    older: Mapping[str, Any], newer: Mapping[str, Any]
) -> Mapping[str, Any]:
    ret: Dict[str, Any] = {}
    for key, val in older.items():
        if key not in newer:
            # keep older key/values
            ret[key] = val
        else:
            # key is present in both newer and older
            new_val = newer[key]
            if isinstance(new_val, Mapping) and isinstance(val, Mapping):
                # merge nested dictionaries
                ret[key] = _merge_user_configs(val, new_val)
            else:
                # for non-dicts newer overrides older
                ret[key] = new_val
    for key in newer.keys() - older.keys():
        # Add keys/values from newer that absent in older
        ret[key] = newer[key]
    return ret


def _check_sections(
    config: Mapping[str, Any],
    valid_names: Set[str],
    filename: Union[str, "os.PathLike[str]"],
) -> None:
    extra_sections = config.keys() - valid_names
    if extra_sections:
        raise ConfigError(
            f"{filename}: unsupported config sections: {extra_sections!r}"
        )
    for name in valid_names:
        section = config.get(name, {})
        if not isinstance(section, dict):
            raise ConfigError(
                f"{filename}: {name!r} should be a section, got {section!r}"
            )


def _check_item(
    val: Any,
    validator: Any,
    full_name: str,
    filename: Union[str, "os.PathLike[str]"],
) -> None:
    if isinstance(validator, tuple):
        container_type, item_type = validator
        if not isinstance(val, container_type):
            raise ConfigError(
                f"{filename}: invalid type for {full_name}, "
                f"{container_type.__name__} is expected"
            )
        for num, i in enumerate(val):
            _check_item(i, item_type, f"{full_name}[{num}]", filename)
    else:
        assert isinstance(validator, type) and issubclass(
            validator, (bool, numbers.Real, numbers.Integral, str)
        )
        # validator for integer types should be numbers.Real or numbers.Integral,
        # not int or float
        if not isinstance(val, validator):
            raise ConfigError(
                f"{filename}: invalid type for {full_name}, "
                f"{validator.__name__} is expected"
            )


def _check_section(
    config: Mapping[str, Any],
    section: str,
    params: Mapping[str, _ParamType],
    filename: Union[str, "os.PathLike[str]"],
) -> None:
    sec = config.get(section)
    if sec is None:
        return
    diff = sec.keys() - params.keys()
    if diff:
        diff_str = ", ".join(f"{section}.{name}" for name in sorted(diff))
        raise ConfigError(f"{filename}: unknown parameters {diff_str}")
    for name, validator in params.items():
        val = sec.get(name)
        if val is None:
            continue
        _check_item(val, validator, f"{section}.{name}", filename)


def _validate_user_config(
    plugin_manager: PluginManager,
    config: Mapping[str, Any],
    filename: Union[str, "os.PathLike[str]"],
    allow_cluster_name: bool = False,
    allow_org_name: bool = False,
) -> None:
    # This was a hard decision.
    # Config structure should be validated to generate meaningful error messages.
    #
    # API should do it but API don't use user config itself, the config is entirely
    # for CLI needs.
    #
    # Since currently CLI is the only API client that reads user config data, API
    # validates it.
    if not allow_cluster_name:
        if "cluster-name" in config.get("job", {}):
            raise ConfigError(
                f"{filename}: cluster name is not allowed in global user "
                f"config file, use 'neuro config switch-cluster' for "
                f"changing the default cluster name"
            )
    if not allow_org_name:
        if "org-name" in config.get("job", {}):
            raise ConfigError(
                f"{filename}: org name is not allowed in global user "
                f"config file, use 'neuro config switch-org' for "
                f"changing the default org name"
            )

    config_spec = plugin_manager.config._get_spec(
        ConfigScope.GLOBAL if not allow_cluster_name else ConfigScope.ALL
    )

    # Alias section uses different validation
    _check_sections(config, set(config_spec.keys()) | {"alias"}, filename)
    for section_name, section_validator in config_spec.items():
        _check_section(config, section_name, section_validator, filename)
    aliases = config.get("alias", {})
    for key, value in aliases.items():
        # check keys and values
        if not CMD_RE.fullmatch(key):
            raise ConfigError(f"{filename}: invalid alias name {key}")
        if not isinstance(value, dict):
            raise ConfigError(
                f"{filename}: invalid alias command type {type(value)}, "
                "run 'neuro help aliases' for getting info about specifying "
                "aliases in config files"
            )
        _validate_alias(key, value, filename)


def _validate_alias(
    key: str, value: Dict[str, str], filename: Union[str, "os.PathLike[str]"]
) -> None:
    # TODO: add validation for both internal and external aliases
    pass


def _load_file(
    plugin_manager: PluginManager,
    filename: Path,
    *,
    allow_cluster_name: bool,
    allow_org_name: bool,
) -> Mapping[str, Any]:
    try:
        config = toml.load(filename)
    except ValueError as exc:
        raise ConfigError(f"{filename}: {exc}")
    _validate_user_config(
        plugin_manager,
        config,
        filename,
        allow_cluster_name=allow_cluster_name,
        allow_org_name=allow_org_name,
    )
    return config


def _load_schema(db: sqlite3.Connection) -> Dict[str, str]:
    cur = db.cursor()
    schema = {}
    cur.execute("SELECT type, name, sql from sqlite_master")
    for type, name, sql in cur:
        if type not in ("table", "index"):
            continue
        if name.startswith("sqlite"):
            # internal object
            continue
        schema[name] = sql
    return schema


def _check_db(db: sqlite3.Connection) -> None:
    schema = _load_schema(db)
    for name, sql in SCHEMA.items():
        if name not in schema:
            raise ConfigError(MALFORMED_CONFIG_MSG)
        if sql != schema[name]:
            raise ConfigError(MALFORMED_CONFIG_MSG)


def _init_db_maybe(db: sqlite3.Connection) -> None:
    # create schema for empty database if needed
    schema = _load_schema(db)
    cur = db.cursor()
    for name, sql in SCHEMA.items():
        if name not in schema:
            cur.execute(sql)



from decimal import Decimal

import aiohttp
import pytest
from aiohttp import web
from yarl import URL

from neuro_sdk import AuthError, Cluster, Preset
from neuro_sdk._login import _AuthConfig
from neuro_sdk._server_cfg import _ServerConfig, get_server_config

from tests import _TestClientFactory


async def test_get_server_config_no_clusters(
    aiohttp_client: _TestClientFactory,
) -> None:
    auth_url = "https://dev-neuro.auth0.com/authorize"
    token_url = "https://dev-neuro.auth0.com/oauth/token"
    logout_url = "https://dev-neuro.auth0.com/v2/logout"
    client_id = "this_is_client_id"
    audience = "https://platform.dev.neu.ro"
    headless_callback_url = "https://dev.neu.ro/oauth/show-code"
    callback_urls = [
        "http://127.0.0.1:54540",
        "http://127.0.0.1:54541",
        "http://127.0.0.1:54542",
    ]
    success_redirect_url = "https://platform.neu.ro"
    JSON = {
        "auth_url": auth_url,
        "token_url": token_url,
        "logout_url": logout_url,
        "client_id": client_id,
        "audience": audience,
        "callback_urls": callback_urls,
        "success_redirect_url": success_redirect_url,
        "headless_callback_url": headless_callback_url,
    }

    async def handler(request: web.Request) -> web.Response:
        assert "Authorization" not in request.headers
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    config = await get_server_config(client.session, client.make_url("/"))
    assert config == _ServerConfig(
        auth_config=_AuthConfig(
            auth_url=URL(auth_url),
            token_url=URL(token_url),
            logout_url=URL(logout_url),
            client_id=client_id,
            audience=audience,
            headless_callback_url=URL(headless_callback_url),
            callback_urls=tuple(URL(u) for u in callback_urls),
            success_redirect_url=URL(success_redirect_url),
        ),
        clusters={},
        admin_url=None,
    )


async def test_get_server_config_no_callback_urls(
    aiohttp_client: _TestClientFactory,
) -> None:

    auth_url = "https://dev-neuro.auth0.com/authorize"
    token_url = "https://dev-neuro.auth0.com/oauth/token"
    logout_url = "https://dev-neuro.auth0.com/v2/logout"
    client_id = "this_is_client_id"
    audience = "https://platform.dev.neu.ro"
    headless_callback_url = "https://dev.neu.ro/oauth/show-code"
    success_redirect_url = "https://platform.neu.ro"
    JSON = {
        "auth_url": auth_url,
        "token_url": token_url,
        "logout_url": logout_url,
        "client_id": client_id,
        "audience": audience,
        "headless_callback_url": headless_callback_url,
        "success_redirect_url": success_redirect_url,
    }

    async def handler(request: web.Request) -> web.Response:
        assert "Authorization" not in request.headers
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    config = await get_server_config(client.session, client.make_url("/"))
    assert config == _ServerConfig(
        auth_config=_AuthConfig(
            auth_url=URL(auth_url),
            token_url=URL(token_url),
            logout_url=URL(logout_url),
            client_id=client_id,
            audience=audience,
            headless_callback_url=URL(headless_callback_url),
            success_redirect_url=URL(success_redirect_url),
        ),
        clusters={},
        admin_url=None,
    )


async def test_get_server_config_with_token(
    aiohttp_client: _TestClientFactory,
) -> None:

    admin_url = "https://admin-dev.neu.ro"
    registry_url = "https://registry.dev.neu.ro"
    storage_url = "https://storage.dev.neu.ro"
    users_url = "https://dev.neu.ro/users"
    monitoring_url = "https://dev.neu.ro/monitoring"
    secrets_url = "https://dev.neu.ro/secrets"
    disks_url = "https://dev.neu.ro/disks"
    buckets_url = "https://dev.neu.ro/buckets"
    auth_url = "https://dev-neuro.auth0.com/authorize"
    token_url = "https://dev-neuro.auth0.com/oauth/token"
    logout_url = "https://dev-neuro.auth0.com/v2/logout"
    client_id = "this_is_client_id"
    audience = "https://platform.dev.neu.ro"
    headless_callback_url = "https://dev.neu.ro/oauth/show-code"
    success_redirect_url = "https://platform.neu.ro"
    JSON = {
        "authorized": True,
        "admin_url": admin_url,
        "auth_url": auth_url,
        "token_url": token_url,
        "logout_url": logout_url,
        "client_id": client_id,
        "audience": audience,
        "headless_callback_url": headless_callback_url,
        "success_redirect_url": success_redirect_url,
        "clusters": [
            {
                "name": "default",
                "monitoring_url": monitoring_url,
                "storage_url": storage_url,
                "registry_url": registry_url,
                "users_url": users_url,
                "secrets_url": secrets_url,
                "disks_url": disks_url,
                "buckets_url": buckets_url,
                "resource_presets": [
                    {
                        "name": "gpu-small",
                        "credits_per_hour": "10",
                        "cpu": 7,
                        "memory_mb": 30 * 1024,
                        "gpu": 1,
                        "gpu_model": "nvidia-tesla-k80",
                    },
                    {
                        "name": "gpu-large",
                        "credits_per_hour": "10",
                        "cpu": 7,
                        "memory_mb": 60 * 1024,
                        "gpu": 1,
                        "gpu_model": "nvidia-tesla-v100",
                    },
                    {
                        "name": "cpu-small",
                        "credits_per_hour": "10",
                        "cpu": 2,
                        "memory_mb": 2 * 1024,
                    },
                    {
                        "name": "cpu-large",
                        "credits_per_hour": "10",
                        "cpu": 3,
                        "memory_mb": 14 * 1024,
                    },
                    {
                        "name": "cpu-large-p",
                        "credits_per_hour": "10",
                        "cpu": 3,
                        "memory_mb": 14 * 1024,
                        "scheduler_enabled": True,
                        "preemptible_node": True,
                    },
                ],
            }
        ],
    }

    async def handler(request: web.Request) -> web.Response:
        assert request.headers["Authorization"] == "Bearer bananatoken"
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    config = await get_server_config(
        client.session, client.make_url("/"), token="bananatoken"
    )
    assert config == _ServerConfig(
        auth_config=_AuthConfig(
            auth_url=URL(auth_url),
            token_url=URL(token_url),
            logout_url=URL(logout_url),
            client_id=client_id,
            audience=audience,
            headless_callback_url=URL(headless_callback_url),
            success_redirect_url=URL(success_redirect_url),
        ),
        admin_url=URL(admin_url),
        clusters={
            "default": Cluster(
                registry_url=URL(registry_url),
                storage_url=URL(storage_url),
                users_url=URL(users_url),
                monitoring_url=URL(monitoring_url),
                secrets_url=URL(secrets_url),
                disks_url=URL(disks_url),
                buckets_url=URL(buckets_url),
                presets={
                    "gpu-small": Preset(
                        credits_per_hour=Decimal("10"),
                        cpu=7,
                        memory_mb=30 * 1024,
                        gpu=1,
                        gpu_model="nvidia-tesla-k80",
                    ),
                    "gpu-large": Preset(
                        credits_per_hour=Decimal("10"),
                        cpu=7,
                        memory_mb=60 * 1024,
                        gpu=1,
                        gpu_model="nvidia-tesla-v100",
                    ),
                    "cpu-small": Preset(
                        credits_per_hour=Decimal("10"), cpu=2, memory_mb=2 * 1024
                    ),
                    "cpu-large": Preset(
                        credits_per_hour=Decimal("10"), cpu=3, memory_mb=14 * 1024
                    ),
                    "cpu-large-p": Preset(
                        credits_per_hour=Decimal("10"),
                        cpu=3,
                        memory_mb=14 * 1024,
                        scheduler_enabled=True,
                        preemptible_node=True,
                    ),
                },
                name="default",
                orgs=[None],
            )
        },
    )


async def test_get_server_config_with_token_no_clusters(
    aiohttp_client: _TestClientFactory,
) -> None:
    auth_url = "https://dev-neuro.auth0.com/authorize"
    token_url = "https://dev-neuro.auth0.com/oauth/token"
    logout_url = "https://dev-neuro.auth0.com/v2/logout"
    client_id = "this_is_client_id"
    audience = "https://platform.dev.neu.ro"
    headless_callback_url = "https://dev.neu.ro/oauth/show-code"
    callback_urls = [
        "http://127.0.0.1:54540",
        "http://127.0.0.1:54541",
        "http://127.0.0.1:54542",
    ]
    success_redirect_url = "https://platform.neu.ro"
    JSON = {
        "authorized": True,
        "auth_url": auth_url,
        "token_url": token_url,
        "logout_url": logout_url,
        "client_id": client_id,
        "audience": audience,
        "callback_urls": callback_urls,
        "success_redirect_url": success_redirect_url,
        "headless_callback_url": headless_callback_url,
    }

    async def handler(request: web.Request) -> web.Response:
        assert request.headers["Authorization"] == "Bearer bananatoken"
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    config = await get_server_config(
        client.session, client.make_url("/"), token="bananatoken"
    )
    assert config == _ServerConfig(
        auth_config=_AuthConfig(
            auth_url=URL(auth_url),
            token_url=URL(token_url),
            logout_url=URL(logout_url),
            client_id=client_id,
            audience=audience,
            headless_callback_url=URL(headless_callback_url),
            callback_urls=tuple(URL(u) for u in callback_urls),
            success_redirect_url=URL(success_redirect_url),
        ),
        clusters={},
        admin_url=None,
    )


async def test_get_server_config_unauthorized(
    aiohttp_client: _TestClientFactory,
) -> None:
    auth_url = "https://dev-neuro.auth0.com/authorize"
    token_url = "https://dev-neuro.auth0.com/oauth/token"
    logout_url = "https://dev-neuro.auth0.com/v2/logout"
    client_id = "this_is_client_id"
    audience = "https://platform.dev.neu.ro"
    headless_callback_url = "https://dev.neu.ro/oauth/show-code"
    callback_urls = [
        "http://127.0.0.1:54540",
        "http://127.0.0.1:54541",
        "http://127.0.0.1:54542",
    ]
    success_redirect_url = "https://platform.neu.ro"
    JSON = {
        "auth_url": auth_url,
        "token_url": token_url,
        "logout_url": logout_url,
        "client_id": client_id,
        "audience": audience,
        "callback_urls": callback_urls,
        "success_redirect_url": success_redirect_url,
        "headless_callback_url": headless_callback_url,
    }

    async def handler(request: web.Request) -> web.Response:
        assert request.headers["Authorization"] == "Bearer bananatoken"
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    with pytest.raises(AuthError, match="Cannot authorize user"):
        await get_server_config(
            client.session, client.make_url("/"), token="bananatoken"
        )


async def test_get_server_config__fail(aiohttp_client: _TestClientFactory) -> None:
    async def handler(request: web.Request) -> web.Response:
        raise aiohttp.web.HTTPInternalServerError(reason="unexpected server error")

    app = web.Application()
    app.router.add_get("/config", handler)
    client = await aiohttp_client(app)

    with pytest.raises(RuntimeError, match="Unable to get server configuration: 500"):
        await get_server_config(client.session, client.make_url("/"))



from datetime import datetime, timezone
from decimal import Decimal
from typing import AsyncIterator, Callable

import pytest
from aiohttp import web
from yarl import URL

from neuro_sdk import Action, Client, Permission, Quota, ResourceNotFound

from tests import _TestServerFactory

_MakeClient = Callable[..., Client]


@pytest.fixture()
async def mocked_share_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    async def handler(request: web.Request) -> web.Response:
        data = await request.json()
        assert data[0]["action"] in [item.value for item in Action]
        data[0]["action"] = Action.MANAGE.value
        return web.json_response(data, status=web.HTTPCreated.status_code)

    app = web.Application()
    app.router.add_post("/users/bill/permissions", handler)
    srv = await aiohttp_server(app)
    client = make_client(srv.make_url("/"))
    yield client
    await client.close()


@pytest.fixture()
async def mocked_revoke_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    async def handler(request: web.Request) -> web.Response:
        assert "uri" in request.query
        raise web.HTTPNoContent()

    app = web.Application()
    app.router.add_delete("/users/bill/permissions", handler)
    srv = await aiohttp_server(app)
    client = make_client(srv.make_url("/"))
    yield client
    await client.close()


@pytest.fixture()
async def mocked_add_role_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    async def handler(request: web.Request) -> web.Response:
        data = await request.json()
        assert data["name"].startswith("mycompany/")
        raise web.HTTPCreated()

    app = web.Application()
    app.router.add_post("/users", handler)
    srv = await aiohttp_server(app)
    client = make_client(srv.make_url("/"))
    yield client
    await client.close()


@pytest.fixture()
async def mocked_remove_role_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    async def handler(request: web.Request) -> web.Response:
        assert request.match_info["name"].startswith("mycompany:")
        raise web.HTTPNoContent()

    app = web.Application()
    app.router.add_delete("/users/{name}", handler)
    srv = await aiohttp_server(app)
    client = make_client(srv.make_url("/"))
    yield client
    await client.close()


@pytest.fixture()
async def mocked_get_quota_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    date = datetime.now(timezone.utc)

    async def handle_get_cluster_user(request: web.Request) -> web.StreamResponse:
        data = {
            "user_name": "denis",
            "role": "admin",
            "user_info": {
                "first_name": "denis",
                "last_name": "admin",
                "email": "denis@domain.name",
                "created_at": date.isoformat(),
            },
            "balance": {
                "credits": "500",
                "spent_credits": "10",
            },
            "quota": {"total_running_jobs": 10},
        }
        return web.json_response(data)

    app = web.Application()
    app.router.add_get(
        "/apis/admin/v1/clusters/{cluster_name}/users/{username}",
        handle_get_cluster_user,
    )

    srv = await aiohttp_server(app)

    client = make_client(srv.make_url("/api/v1"))
    yield client
    await client.close()


@pytest.fixture()
async def mocked_get_subroles_client(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> AsyncIterator[Client]:
    async def handle_get_subroles(request: web.Request) -> web.StreamResponse:
        username = request.match_info["username"]
        data = {
            "subroles": [f"{username}/sub1", f"{username}/sub2", f"{username}/sub3"]
        }
        return web.json_response(data)

    app = web.Application()
    app.router.add_get(
        "/api/v1/users/{username}/subroles",
        handle_get_subroles,
    )

    srv = await aiohttp_server(app)

    client = make_client(srv.make_url("/api/v1"))
    yield client
    await client.close()


class TestUsers:
    async def test_get_quota(self, mocked_get_quota_client: Client) -> None:
        res = await mocked_get_quota_client.users.get_quota()
        assert res == Quota(credits=Decimal("500"), total_running_jobs=10)

    async def test_get_quota_adminless(self, make_client: _MakeClient) -> None:
        async with make_client("https://dev.example.com", admin_url="") as client:
            quota = await client.users.get_quota()
            assert quota.credits is None
            assert quota.total_running_jobs is None

    async def test_share_unknown_user(self, mocked_share_client: Client) -> None:
        with pytest.raises(ResourceNotFound):
            await mocked_share_client.users.share(
                user="not-exists",
                permission=Permission(URL("storage://bob/resource"), Action.READ),
            )

    async def test_share_invalid_name(self, mocked_share_client: Client) -> None:
        with pytest.raises(ValueError):
            await mocked_share_client.users.share(
                user="mycompany/team:role",
                permission=Permission(URL("storage://bob/resource"), Action.READ),
            )

    async def test_correct_share(self, mocked_share_client: Client) -> None:
        ret = await mocked_share_client.users.share(
            user="bill",
            permission=Permission(URL("storage://bob/resource"), Action.READ),
        )
        assert ret == Permission(URL("storage://bob/resource"), Action.MANAGE)

    async def test_revoke_unknown_user(self, mocked_revoke_client: Client) -> None:
        with pytest.raises(ResourceNotFound):
            await mocked_revoke_client.users.revoke(
                user="not-exists", uri=URL("storage://bob/resource")
            )

    async def test_revoke_invalid_name(self, mocked_revoke_client: Client) -> None:
        with pytest.raises(ValueError):
            await mocked_revoke_client.users.revoke(
                user="mycompany/team:role", uri=URL("storage://bob/resource")
            )

    async def test_correct_revoke(self, mocked_revoke_client: Client) -> None:
        ret = await mocked_revoke_client.users.revoke(
            user="bill", uri=URL("storage://bob/resource")
        )
        assert ret is None  # at this moment no result

    async def test_add_role(self, mocked_add_role_client: Client) -> None:
        ret = await mocked_add_role_client.users.add("mycompany/team/role")
        assert ret is None  # at this moment no result

    async def test_remove_role(self, mocked_remove_role_client: Client) -> None:
        ret = await mocked_remove_role_client.users.remove("mycompany/team/role")
        assert ret is None  # at this moment no result

    async def test_get_subroles(self, mocked_get_subroles_client: Client) -> None:
        res = await mocked_get_subroles_client.users.get_subroles("test")
        assert set(res) == {"test/sub1", "test/sub2", "test/sub3"}

    async def test_remove_role_invalid_name(
        self, mocked_remove_role_client: Client
    ) -> None:
        with pytest.raises(ValueError):
            await mocked_remove_role_client.users.remove("mycompany/team:role")



import logging
from typing import Any

import aiohttp
import pytest

from neuro_sdk._utils import retries


async def test_success(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    for retry in retries("Fails", attempts=5):
        async with retry:
            count += 1

    assert count == 1
    assert caplog.record_tuples == []


async def test_first_fails(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    for retry in retries("Fails", attempts=5):
        async with retry:
            count += 1
            if count == 1:
                raise aiohttp.ClientError("Ouch!")

    assert count == 2
    record = ("neuro_sdk", logging.INFO, "Fails: Ouch!.  Retry...")
    assert caplog.record_tuples == [record]


async def test_two_fail(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    for retry in retries("Fails", attempts=5):
        async with retry:
            count += 1
            if count <= 2:
                raise aiohttp.ClientError("Ouch!")

    assert count == 3
    record = ("neuro_sdk", logging.INFO, "Fails: Ouch!.  Retry...")
    assert caplog.record_tuples == [record] * 2


async def test_all_fail(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    with pytest.raises(aiohttp.ClientError):
        for retry in retries("Fails", attempts=3):
            async with retry:
                count += 1
                raise aiohttp.ClientError("Ouch!")

    assert count == 3
    record = ("neuro_sdk", logging.INFO, "Fails: Ouch!.  Retry...")
    assert caplog.record_tuples == [record] * 2


async def test_reset(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    for retry in retries("Fails", attempts=3):
        async with retry:
            count += 1
            if count in (3, 5):
                retry.reset()
            if count <= 6:
                raise aiohttp.ClientError("Ouch!")

    assert count == 7
    record = ("neuro_sdk", logging.INFO, "Fails: Ouch!.  Retry...")
    assert caplog.record_tuples == [record] * 6


async def test_unexpected_error(caplog: Any) -> None:
    caplog.set_level(logging.INFO)
    count = 0
    with pytest.raises(ZeroDivisionError):
        for retry in retries("Fails", attempts=5):
            async with retry:
                count += 1
                1 / 0

    assert count == 1
    assert caplog.record_tuples == []



from yarl import URL

import neuro_sdk


def test_public_version() -> None:
    assert "__version__" in neuro_sdk.__all__


def test_module_for_public_names() -> None:
    for name in neuro_sdk.__all__:
        obj = getattr(neuro_sdk, name)
        if isinstance(obj, URL):
            # Default API url
            continue
        if hasattr(obj, "__module__"):
            assert obj.__module__ in (
                "neuro_sdk",
                # objects from typing are public type hint aliases,
                # e.g. Callable[...]
                "typing",
                # We re-export entities from admin client
                "neuro_admin_client.entities",
            ), f"{obj}.__module__ == {obj.__module__}, expected neuro_sdk"



import asyncio
import base64
import json
import os
import ssl
import sys
from pathlib import Path
from typing import Any, Awaitable, Callable, List, Mapping, Optional

import aiohttp
import certifi
from yarl import URL

from ._client import Client
from ._config import _ConfigData, _load, _load_recovery_data, _load_user_config, _save
from ._core import DEFAULT_TIMEOUT
from ._errors import ConfigError
from ._login import (
    AuthNegotiator,
    AuthTokenClient,
    HeadlessNegotiator,
    _AuthToken,
    create_standalone_token,
    logout_from_browser,
)
from ._plugins import PluginManager
from ._rewrite import rewrite_module
from ._server_cfg import _ServerConfig, get_server_config
from ._tracing import _make_trace_config
from ._utils import _ContextManager

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points

DEFAULT_CONFIG_PATH = "~/.neuro"
CONFIG_ENV_NAME = "NEUROMATION_CONFIG"
PASS_CONFIG_ENV_NAME = "NEURO_PASSED_CONFIG"
DEFAULT_API_URL = URL("https://staging.neu.ro/api/v1")


def _make_session(
    timeout: aiohttp.ClientTimeout, trace_configs: Optional[List[aiohttp.TraceConfig]]
) -> _ContextManager[aiohttp.ClientSession]:
    return _ContextManager[aiohttp.ClientSession](
        __make_session(timeout, trace_configs)
    )


async def __make_session(
    timeout: aiohttp.ClientTimeout, trace_configs: Optional[List[aiohttp.TraceConfig]]
) -> aiohttp.ClientSession:
    from . import __version__

    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
    ssl_context.load_verify_locations(capath=certifi.where())
    connector = aiohttp.TCPConnector(ssl=ssl_context)
    return aiohttp.ClientSession(
        timeout=timeout,
        connector=connector,
        trace_configs=trace_configs,
        headers={"User-Agent": f"NeuroCLI/{__version__} ({sys.platform})"},
    )


@rewrite_module
class Factory:
    def __init__(
        self,
        path: Optional[Path] = None,
        trace_configs: Optional[List[aiohttp.TraceConfig]] = None,
        trace_id: Optional[str] = None,
        trace_sampled: Optional[bool] = None,
    ) -> None:
        if path is None:
            path = Path(os.environ.get(CONFIG_ENV_NAME, DEFAULT_CONFIG_PATH))
        self._path = path.expanduser()
        self._trace_configs = [_make_trace_config()]
        if trace_configs:
            self._trace_configs += trace_configs
        self._trace_id = trace_id
        self._trace_sampled = trace_sampled
        self._plugin_manager = PluginManager()
        for entry_point in entry_points(group="neuro_api"):
            entry_point.load()(self._plugin_manager)

    @property
    def path(self) -> Path:
        return self._path

    @property
    def is_config_present(self) -> bool:
        return (self._path / "db").exists()

    async def get(self, *, timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT) -> Client:
        if not self.is_config_present and PASS_CONFIG_ENV_NAME in os.environ:
            await self.login_with_passed_config(timeout=timeout)
        try:
            return await self._get(timeout=timeout)
        except ConfigError as initial_error:
            try:
                await self._try_recover_config(timeout)
            except asyncio.CancelledError:
                raise
            except Exception:
                raise initial_error
            return await self._get(timeout=timeout)

    async def _get(self, *, timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT) -> Client:
        session = await _make_session(timeout, self._trace_configs)
        try:
            client = Client._create(
                session,
                self._path,
                self._trace_id,
                self._trace_sampled,
                self._plugin_manager,
            )
            await client.config.check_server()
        except (asyncio.CancelledError, Exception):
            await session.close()
            raise
        else:
            return client

    async def _try_recover_config(
        self, timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT
    ) -> None:
        recovery_data = _load_recovery_data(self._path)
        async with _make_session(timeout, self._trace_configs) as session:
            config_unauthorized = await get_server_config(session, recovery_data.url)
            old_token = _AuthToken.create("", 0, recovery_data.refresh_token)
            async with AuthTokenClient(
                session,
                url=config_unauthorized.auth_config.token_url,
                client_id=config_unauthorized.auth_config.client_id,
            ) as token_client:
                fresh_token = await token_client.refresh(old_token)
            config_authorized = await get_server_config(
                session, recovery_data.url, token=fresh_token.token
            )
            config = self._gen_config(config_authorized, fresh_token, recovery_data.url)
        await self.logout()  # Drop old broken config
        self._save(config)

        client = await self.get(timeout=timeout)
        await client.config.switch_cluster(recovery_data.cluster_name)
        await client.config.switch_org(recovery_data.org_name)
        await client.close()

    async def login(
        self,
        show_browser_cb: Callable[[URL], Awaitable[None]],
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            config_unauthorized = await get_server_config(session, url)

            if config_unauthorized.clusters:
                config_authorized = config_unauthorized
                auth_token = create_standalone_token()
            else:
                negotiator = AuthNegotiator(
                    session, config_unauthorized.auth_config, show_browser_cb
                )
                auth_token = await negotiator.get_token()

                config_authorized = await get_server_config(
                    session, url, token=auth_token.token
                )
            config = self._gen_config(config_authorized, auth_token, url)
        self._save(config)

    async def login_headless(
        self,
        get_auth_code_cb: Callable[[URL], Awaitable[str]],
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            config_unauthorized = await get_server_config(session, url)
            negotiator = HeadlessNegotiator(
                session, config_unauthorized.auth_config, get_auth_code_cb
            )
            auth_token = await negotiator.get_token()

            config_authorized = await get_server_config(
                session, url, token=auth_token.token
            )
        config = self._gen_config(config_authorized, auth_token, url)
        self._save(config)

    async def login_with_token(
        self,
        token: str,
        *,
        url: URL = DEFAULT_API_URL,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        config_file = self._path / "db"
        if config_file.exists():
            raise ConfigError(f"Config at {self._path} already exists. Please logout")
        async with _make_session(timeout, self._trace_configs) as session:
            server_config = await get_server_config(session, url, token=token)
        config = self._gen_config(
            server_config, _AuthToken.create_non_expiring(token), url
        )
        self._save(config)

    async def login_with_passed_config(
        self,
        config_data: Optional[str] = None,
        *,
        timeout: aiohttp.ClientTimeout = DEFAULT_TIMEOUT,
    ) -> None:
        if config_data is None:
            try:
                config_data = os.environ[PASS_CONFIG_ENV_NAME]
            except KeyError:
                raise ConfigError(
                    f"Config env variable {PASS_CONFIG_ENV_NAME} " "is not present"
                )
        try:
            data = json.loads(base64.b64decode(config_data).decode())
            token = data["token"]
            cluster = data["cluster"]
            url = URL(data["url"])
        except (ValueError, KeyError):
            raise ConfigError(f"Data in passed config is malformed: {config_data}")
        await self.login_with_token(token, url=url, timeout=timeout)
        client = await self.get(timeout=timeout)

        await client.config.switch_cluster(cluster)
        await client.close()

    def _gen_config(
        self, server_config: _ServerConfig, token: _AuthToken, url: URL
    ) -> _ConfigData:
        from . import __version__

        if server_config.clusters:
            cluster_name = next(iter(server_config.clusters))
            org_name = server_config.clusters[cluster_name].orgs[0]
        else:
            cluster_name = None
            org_name = None
        config = _ConfigData(
            auth_config=server_config.auth_config,
            auth_token=token,
            url=url,
            admin_url=server_config.admin_url,
            version=__version__,
            cluster_name=cluster_name,
            org_name=org_name,
            clusters=server_config.clusters,
        )
        return config

    async def logout(
        self,
        show_browser_cb: Callable[[URL], Awaitable[None]] = None,
    ) -> None:
        if show_browser_cb is not None:
            try:
                old_config = _load(self._path)
            except ConfigError:
                pass  # Do not try to logout from auth0 if config is broken
            else:
                await logout_from_browser(old_config.auth_config, show_browser_cb)

        files = ["db", "db-wal", "db-shm"]
        for name in files:
            f = self._path / name
            if f.exists():
                f.unlink()
        if self._path.is_file():
            # Old-styled single file config from 2019
            self._path.unlink()
        else:
            try:
                self._path.rmdir()
            except OSError:
                # Directory Not Empty or Not A Directory
                pass

    async def load_user_config(self) -> Mapping[str, Any]:
        return _load_user_config(self._plugin_manager, self._path)

    def _save(self, config: _ConfigData) -> None:
        _save(config, self._path, False)



import abc
import asyncio
import base64
import errno
import hashlib
import secrets
import time
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    List,
    Optional,
    Sequence,
    Type,
    cast,
)
from urllib.parse import urlencode

import aiohttp
from aiohttp import ClientResponseError
from aiohttp.web import (
    Application,
    AppRunner,
    HTTPBadRequest,
    HTTPForbidden,
    HTTPFound,
    HTTPUnauthorized,
    Request,
    Response,
    TCPSite,
)
from jose import JWTError, jwt
from yarl import URL

from ._errors import AuthError


def urlsafe_unpadded_b64encode(payload: bytes) -> str:
    return base64.urlsafe_b64encode(payload).decode().rstrip("=")


# Used only for standalone platform deploymens
JWT_STANDALONE_SECRET = "neuro"

JWT_IDENTITY_CLAIM = "https://platform.neuromation.io/user"
JWT_IDENTITY_CLAIM_OPTIONS = ("identity", JWT_IDENTITY_CLAIM)


class AuthCode:
    def __init__(self, callback_url: Optional[URL] = None) -> None:
        self._future: asyncio.Future[str] = asyncio.Future()

        self._verifier = urlsafe_unpadded_b64encode(secrets.token_bytes(32))
        digest = hashlib.sha256(self._verifier.encode()).digest()
        self._challenge = urlsafe_unpadded_b64encode(digest)
        self._challenge_method = "S256"

        self._callback_url = callback_url

    @property
    def verifier(self) -> str:
        return self._verifier

    @property
    def challenge(self) -> str:
        return self._challenge

    @property
    def challenge_method(self) -> str:
        return self._challenge_method

    def set_value(self, value: str) -> None:
        if not self._future.cancelled():
            self._future.set_result(value)

    @property
    def callback_url(self) -> URL:
        assert self._callback_url
        return self._callback_url

    @callback_url.setter
    def callback_url(self, value: URL) -> None:
        self._callback_url = value

    def set_exception(self, exc: Exception) -> None:
        if not self._future.cancelled():
            self._future.set_exception(exc)

    def cancel(self) -> None:
        self._future.cancel()

    async def wait(self, timeout_s: float = 60.0) -> str:
        try:
            await asyncio.wait_for(self._future, timeout_s)
        except (asyncio.TimeoutError, asyncio.CancelledError):
            raise AuthError("failed to get an authorization code")
        return self._future.result()


class AuthCodeCallbackClient(abc.ABC):
    def __init__(self, url: URL, client_id: str, audience: str) -> None:
        self._url = url
        self._client_id = client_id
        self._audience = audience

    async def request(self, code: AuthCode) -> AuthCode:
        url = self._url.with_query(
            response_type="code",
            code_challenge=code.challenge,
            code_challenge_method=code.challenge_method,
            client_id=self._client_id,
            redirect_uri=str(code.callback_url),
            scope="offline_access",
            audience=self._audience,
        )
        await self._request(url, code)
        await code.wait()
        return code

    @abc.abstractmethod
    async def _request(self, url: URL, code: AuthCode) -> None:
        pass


class WebBrowserAuthCodeCallbackClient(AuthCodeCallbackClient):
    def __init__(
        self,
        url: URL,
        client_id: str,
        audience: str,
        show_browser_cb: Callable[[URL], Awaitable[None]],
    ) -> None:
        super().__init__(url=url, client_id=client_id, audience=audience)
        self._show_browser_cb = show_browser_cb

    async def _request(self, url: URL, code: AuthCode) -> None:
        await self._show_browser_cb(url)


class HeadlessAuthCodeCallbackClient(AuthCodeCallbackClient):
    def __init__(
        self,
        url: URL,
        client_id: str,
        audience: str,
        get_auth_code_cb: Callable[[URL], Awaitable[str]],
    ) -> None:
        super().__init__(url=url, client_id=client_id, audience=audience)
        self._get_auth_code_cb = get_auth_code_cb

    async def _request(self, url: URL, code: AuthCode) -> None:
        try:
            auth_code = await self._get_auth_code_cb(url)
            assert auth_code
            code.set_value(auth_code)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            code.set_exception(exc)


class AuthCodeCallbackHandler:
    def __init__(self, code: AuthCode, redirect_url: Optional[URL] = None) -> None:
        self._code = code
        self._redirect_url = redirect_url

    async def handle(self, request: Request) -> Response:
        if "error" in request.query:
            await self._handle_error(request)

        code = request.query.get("code")

        if not code:
            self._code.cancel()
            raise HTTPBadRequest(text="The 'code' query parameter is missing.")

        self._code.set_value(code)

        if self._redirect_url:
            raise HTTPFound(self._redirect_url)
        return Response(text="OK")

    async def _handle_error(self, request: Request) -> None:
        error = request.query["error"]
        description = request.query.get("error_description", "")

        exc_factory: Type[Exception]
        if error == "unauthorized":
            exc_factory = HTTPUnauthorized
        elif error == "access_denied":
            exc_factory = HTTPForbidden
        else:
            exc_factory = HTTPBadRequest

        self._code.set_exception(AuthError(description))
        raise exc_factory(text=description)


def create_auth_code_app(
    code: AuthCode, redirect_url: Optional[URL] = None
) -> Application:
    app = Application()
    handler = AuthCodeCallbackHandler(code, redirect_url=redirect_url)
    app.router.add_get("/", handler.handle)
    return app


@asynccontextmanager
async def create_app_server_once(
    app: Application, *, host: str = "127.0.0.1", port: int = 8080
) -> AsyncIterator[URL]:
    runner = AppRunner(app, access_log=None)
    try:
        await runner.setup()
        site = TCPSite(runner, host, port, shutdown_timeout=0.0)
        await site.start()
        yield URL(site.name)
    finally:
        await runner.shutdown()
        await runner.cleanup()


@asynccontextmanager
async def create_app_server(
    app: Application, *, host: str = "127.0.0.1", ports: Sequence[int] = (8080,)
) -> AsyncIterator[URL]:
    for port in ports:
        try:
            async with create_app_server_once(app, host=host, port=port) as url:
                yield url
            return
        except OSError as err:
            if err.errno != errno.EADDRINUSE:
                raise
    else:
        raise RuntimeError("No free ports.")


@dataclass(frozen=True)
class _AuthToken:
    token: str
    expiration_time: float
    refresh_token: str = field(repr=False)

    def is_expired(self, *, now: Optional[float] = None) -> bool:
        if now is None:
            now = time.time()
        return self.expiration_time <= now

    @property
    def username(self) -> str:
        try:
            claims = jwt.get_unverified_claims(self.token)
        except JWTError as e:
            raise ValueError(
                f"Passed string does not contain valid JWT structure."
            ) from e
        for identity_claim in JWT_IDENTITY_CLAIM_OPTIONS:
            if identity_claim in claims:
                return claims[identity_claim]
        raise ValueError("JWT Claims structure is not correct.")

    @classmethod
    def create(
        cls,
        token: str,
        expires_in: float,
        refresh_token: str,
        expiration_ratio: float = 0.75,
        *,
        now: Optional[float] = None,
    ) -> "_AuthToken":
        if now is None:
            now = time.time()
        expiration_time = now + expires_in * expiration_ratio
        return cls(
            token=token,
            expiration_time=expiration_time,
            refresh_token=refresh_token,
        )

    @classmethod
    def create_non_expiring(cls, token: str) -> "_AuthToken":
        # NOTE: for backward compatibility we assume that manually set token
        # expires in 3 years.
        expires_in = 60 * 60 * 24 * 365 * 3  # 3 years
        return cls.create(token, expires_in=expires_in, refresh_token="")


class AuthTokenClient:
    def __init__(
        self, session: aiohttp.ClientSession, url: URL, client_id: str
    ) -> None:
        self._url = url
        self._client_id = client_id

        self._client = session

    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "AuthTokenClient":
        return self

    async def __aexit__(self, *_: Sequence[Any]) -> None:
        await self.close()

    async def request(self, code: AuthCode) -> _AuthToken:
        payload = dict(
            grant_type="authorization_code",
            code_verifier=code.verifier,
            code=await code.wait(),
            client_id=self._client_id,
            redirect_uri=str(code.callback_url),
        )
        async with self._client.post(
            self._url,
            headers={
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
            },
            data=urlencode(payload),
        ) as resp:
            try:
                resp.raise_for_status()
            except ClientResponseError as exc:
                raise AuthError("failed to get an access token.") from exc
            resp_payload = await resp.json()
            return _AuthToken.create(
                token=resp_payload["access_token"],
                expires_in=resp_payload["expires_in"],
                refresh_token=resp_payload["refresh_token"],
            )

    async def refresh(self, token: _AuthToken) -> _AuthToken:
        payload = dict(
            grant_type="refresh_token",
            refresh_token=token.refresh_token,
            client_id=self._client_id,
        )
        async with self._client.post(
            self._url,
            headers={
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
            },
            data=urlencode(payload),
        ) as resp:
            try:
                resp.raise_for_status()
            except ClientResponseError as exc:
                raise AuthError("failed to get an access token.") from exc
            resp_payload = await resp.json()
            return _AuthToken.create(
                token=resp_payload["access_token"],
                expires_in=resp_payload["expires_in"],
                refresh_token=token.refresh_token,
            )


@dataclass(frozen=True)
class _AuthConfig:
    auth_url: URL
    token_url: URL
    logout_url: URL

    client_id: str
    audience: str

    headless_callback_url: URL

    callback_urls: Sequence[URL] = (
        URL("http://127.0.0.1:54540"),
        URL("http://127.0.0.1:54541"),
        URL("http://127.0.0.1:54542"),
    )

    success_redirect_url: Optional[URL] = None

    @property
    def callback_host(self) -> str:
        return cast(str, self.callback_urls[0].host)

    @property
    def callback_ports(self) -> List[int]:
        return [cast(int, url.port) for url in self.callback_urls]

    @classmethod
    def create(
        cls,
        auth_url: URL,
        token_url: URL,
        logout_url: URL,
        client_id: str,
        audience: str,
        headless_callback_url: URL,
        success_redirect_url: Optional[URL] = None,
        callback_urls: Optional[Sequence[URL]] = None,
    ) -> "_AuthConfig":
        return cls(
            auth_url=auth_url,
            token_url=token_url,
            logout_url=logout_url,
            client_id=client_id,
            audience=audience,
            headless_callback_url=headless_callback_url,
            success_redirect_url=success_redirect_url,
            callback_urls=callback_urls or cls.callback_urls,
        )


class BaseNegotiator(abc.ABC):
    def __init__(self, session: aiohttp.ClientSession, config: _AuthConfig) -> None:
        self._config = config
        self._session = session

    @abc.abstractmethod
    async def get_code(self) -> AuthCode:
        pass

    async def get_token(self) -> _AuthToken:
        async with AuthTokenClient(
            self._session, url=self._config.token_url, client_id=self._config.client_id
        ) as token_client:
            code = await self.get_code()
            return await token_client.request(code)


class AuthNegotiator(BaseNegotiator):
    def __init__(
        self,
        session: aiohttp.ClientSession,
        config: _AuthConfig,
        show_browser_cb: Callable[[URL], Awaitable[None]],
    ) -> None:
        super().__init__(session, config)
        self._show_browser_cb = show_browser_cb

    async def get_code(self) -> AuthCode:
        code = AuthCode()
        app = create_auth_code_app(code, redirect_url=self._config.success_redirect_url)

        async with create_app_server(
            app, host=self._config.callback_host, ports=self._config.callback_ports
        ) as url:
            code.callback_url = url
            code_callback_client = WebBrowserAuthCodeCallbackClient(
                url=self._config.auth_url,
                client_id=self._config.client_id,
                audience=self._config.audience,
                show_browser_cb=self._show_browser_cb,
            )
            return await code_callback_client.request(code)


class HeadlessNegotiator(BaseNegotiator):
    def __init__(
        self,
        session: aiohttp.ClientSession,
        config: _AuthConfig,
        get_auth_code_cb: Callable[[URL], Awaitable[str]],
    ) -> None:
        super().__init__(session, config)
        self._get_auth_code_cb = get_auth_code_cb

    async def get_code(self) -> AuthCode:
        code = AuthCode()
        code.callback_url = self._config.headless_callback_url

        code_callback_client = HeadlessAuthCodeCallbackClient(
            url=self._config.auth_url,
            client_id=self._config.client_id,
            audience=self._config.audience,
            get_auth_code_cb=self._get_auth_code_cb,
        )
        return await code_callback_client.request(code)


async def logout_from_browser(
    config: _AuthConfig, show_browser_cb: Callable[[URL], Awaitable[None]]
) -> None:
    logout_url = config.logout_url.update_query(client_id=config.client_id)
    await show_browser_cb(logout_url)


def create_standalone_token() -> _AuthToken:
    token = jwt.encode({JWT_IDENTITY_CLAIM: "user"}, JWT_STANDALONE_SECRET)
    return _AuthToken.create_non_expiring(token)



from typing import Callable

from aiohttp import web

from neuro_sdk import (
    Bucket,
    BucketCredentials,
    Client,
    Cluster,
    PersistentBucketCredentials,
)

from tests import _TestServerFactory

_MakeClient = Callable[..., Client]


async def test_list(
    aiohttp_server: _TestServerFactory,
    make_client: _MakeClient,
    cluster_config: Cluster,
) -> None:
    async def handler(request: web.Request) -> web.Response:
        return web.json_response(
            [
                {
                    "id": "credentials-1",
                    "owner": "user",
                    "name": None,
                    "read_only": True,
                    "credentials": [
                        {
                            "bucket_id": "test-1",
                            "provider": "aws",
                            "credentials": {
                                "key": "value",
                            },
                        },
                        {
                            "bucket_id": "test-2",
                            "provider": "aws",
                            "credentials": {
                                "key": "value",
                            },
                        },
                    ],
                },
                {
                    "id": "credentials-2",
                    "owner": "user",
                    "name": None,
                    "read_only": False,
                    "credentials": [
                        {
                            "bucket_id": "test-3",
                            "provider": "aws",
                            "credentials": {
                                "key": "value",
                            },
                        },
                        {
                            "bucket_id": "test-4",
                            "provider": "aws",
                            "credentials": {
                                "key": "value",
                            },
                        },
                    ],
                },
            ]
        )

    app = web.Application()
    app.router.add_get("/buckets/persistent_credentials", handler)

    srv = await aiohttp_server(app)

    ret = []

    async with make_client(srv.make_url("/")) as client:
        async with client.buckets.persistent_credentials_list() as it:
            async for s in it:
                ret.append(s)

    assert ret == [
        PersistentBucketCredentials(
            id="credentials-1",
            owner="user",
            cluster_name=cluster_config.name,
            name=None,
            read_only=True,
            credentials=[
                BucketCredentials(
                    bucket_id="test-1",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
                BucketCredentials(
                    bucket_id="test-2",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
            ],
        ),
        PersistentBucketCredentials(
            id="credentials-2",
            owner="user",
            cluster_name=cluster_config.name,
            name=None,
            read_only=False,
            credentials=[
                BucketCredentials(
                    bucket_id="test-3",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
                BucketCredentials(
                    bucket_id="test-4",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
            ],
        ),
    ]


async def test_add(
    aiohttp_server: _TestServerFactory,
    make_client: _MakeClient,
    cluster_config: Cluster,
) -> None:
    async def handler(request: web.Request) -> web.Response:
        data = await request.json()
        assert data == {
            "name": "test-bucket",
            "bucket_ids": ["test-1", "test-2"],
            "read_only": False,
        }
        return web.json_response(
            {
                "id": "credentials-1",
                "owner": "user",
                "name": None,
                "credentials": [
                    {
                        "bucket_id": "test-1",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                    {
                        "bucket_id": "test-2",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                ],
            }
        )

    app = web.Application()
    app.router.add_post("/buckets/persistent_credentials", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        bucket = await client.buckets.persistent_credentials_create(
            name="test-bucket", bucket_ids=["test-1", "test-2"]
        )
        assert bucket == PersistentBucketCredentials(
            id="credentials-1",
            owner="user",
            cluster_name=cluster_config.name,
            name=None,
            read_only=False,
            credentials=[
                BucketCredentials(
                    bucket_id="test-1",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
                BucketCredentials(
                    bucket_id="test-2",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
            ],
        )


async def test_add_readonly(
    aiohttp_server: _TestServerFactory,
    make_client: _MakeClient,
    cluster_config: Cluster,
) -> None:
    async def handler(request: web.Request) -> web.Response:
        data = await request.json()
        assert data == {
            "name": "test-bucket",
            "bucket_ids": ["test-1", "test-2"],
            "read_only": True,
        }
        return web.json_response(
            {
                "id": "credentials-1",
                "owner": "user",
                "name": None,
                "read_only": True,
                "credentials": [
                    {
                        "bucket_id": "test-1",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                    {
                        "bucket_id": "test-2",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                ],
            }
        )

    app = web.Application()
    app.router.add_post("/buckets/persistent_credentials", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        bucket = await client.buckets.persistent_credentials_create(
            name="test-bucket",
            bucket_ids=["test-1", "test-2"],
            read_only=True,
        )
        assert bucket == PersistentBucketCredentials(
            id="credentials-1",
            owner="user",
            cluster_name=cluster_config.name,
            name=None,
            read_only=True,
            credentials=[
                BucketCredentials(
                    bucket_id="test-1",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
                BucketCredentials(
                    bucket_id="test-2",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
            ],
        )


async def test_get(
    aiohttp_server: _TestServerFactory,
    make_client: _MakeClient,
    cluster_config: Cluster,
) -> None:
    async def handler(request: web.Request) -> web.Response:
        assert request.match_info["key"] == "name"
        return web.json_response(
            {
                "id": "credentials-1",
                "owner": "user",
                "name": None,
                "credentials": [
                    {
                        "bucket_id": "test-1",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                    {
                        "bucket_id": "test-2",
                        "provider": "aws",
                        "credentials": {
                            "key": "value",
                        },
                    },
                ],
            }
        )

    app = web.Application()
    app.router.add_get("/buckets/persistent_credentials/{key}", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        bucket = await client.buckets.persistent_credentials_get("name")
        assert bucket == PersistentBucketCredentials(
            id="credentials-1",
            owner="user",
            cluster_name=cluster_config.name,
            name=None,
            read_only=False,
            credentials=[
                BucketCredentials(
                    bucket_id="test-1",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
                BucketCredentials(
                    bucket_id="test-2",
                    provider=Bucket.Provider.AWS,
                    credentials={
                        "key": "value",
                    },
                ),
            ],
        )


async def test_rm(aiohttp_server: _TestServerFactory, make_client: _MakeClient) -> None:
    async def handler(request: web.Request) -> web.Response:
        assert request.match_info["key"] == "name"
        raise web.HTTPNoContent

    app = web.Application()
    app.router.add_delete("/buckets/persistent_credentials/{key}", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        await client.buckets.persistent_credentials_rm("name")
